@@ -29,6 +29,7 @@
 
 #include "support/bind.h"
 #include "support/regex.h"
+#include "support/TempFile.h"
 
 #include <fstream>
 
@@ -40,13 +41,14 @@ namespace lyx {
 
 
 LayoutFile::LayoutFile(string const & fn, string const & cln,
-			   string const & desc, string const & prereq,
-				 bool texclassavail) 
+		       string const & desc, string const & prereq,
+		       string const & category, bool texclassavail) 
 {
 	name_ = fn;
 	latexname_ = cln;
 	description_ = desc;
 	prerequisites_ = prereq;
+	category_ = category;
 	tex_class_avail_ = texclassavail;
 }
 
@@ -82,14 +84,16 @@ bool LayoutFileList::haveClass(string const & classname) const
 
 LayoutFile const & LayoutFileList::operator[](string const & classname) const
 {
-	LASSERT(haveClass(classname), /**/);
+	LATTEST(haveClass(classname));
+	// safe to continue, since we will make an empty LayoutFile
 	return *classmap_[classname];
 }
 
 
 LayoutFile & LayoutFileList::operator[](string const & classname)
 {
-	LASSERT(haveClass(classname), /**/);
+	LATTEST(haveClass(classname));
+	// safe to continue, since we will make an empty LayoutFile
 	return *classmap_[classname];
 }
 
@@ -141,9 +145,13 @@ bool LayoutFileList::read()
 					break;
 				string const prereq = lex.getString();
 				LYXERR(Debug::TCLASS, "Prereq: " << prereq);
+				if (!lex.next()) 
+					break;
+				string const category = lex.getString();
+				LYXERR(Debug::TCLASS, "Category: " << category);
 				// This code is run when we have
 				// fname, clname, desc, prereq, and avail
-				LayoutFile * tmpl = new LayoutFile(fname, clname, desc, prereq, avail);
+				LayoutFile * tmpl = new LayoutFile(fname, clname, desc, prereq, category, avail);
 				if (lyxerr.debugging(Debug::TCLASS)) {
 					// only system layout files are loaded here so no
 					// buffer path is needed.
@@ -177,12 +185,15 @@ std::vector<LayoutFileIndex> LayoutFileList::classList() const
 }
 
 
-void LayoutFileList::reset(LayoutFileIndex const & classname) {
-	LASSERT(haveClass(classname), /**/);
+void LayoutFileList::reset(LayoutFileIndex const & classname)
+{
+	LATTEST(haveClass(classname));
+	// safe to continue, since we will make an empty LayoutFile
 	LayoutFile * tc = classmap_[classname];
 	LayoutFile * tmpl = 
 		new LayoutFile(tc->name(), tc->latexname(), tc->description(),
-		               tc->prerequisites(), tc->isTeXClassAvailable());
+		               tc->prerequisites(), tc->category(),
+			       tc->isTeXClassAvailable());
 	classmap_[classname] = tmpl;
 	delete tc;
 }
@@ -210,9 +221,13 @@ string layoutpost =
 	
 }
 
+
 LayoutFileIndex LayoutFileList::addEmptyClass(string const & textclass)
 {
-	FileName const tempLayout = FileName::tempName();
+	// FIXME This could be simplified a bit to call TextClass::read(string, ReadType).
+
+	TempFile tempfile("basicXXXXXX.layout");
+	FileName const tempLayout = tempfile.name();
 	ofstream ofs(tempLayout.toFilesystemEncoding().c_str());
 	// This writes a very basic class, but it also attempts to include 
 	// stdclass.inc. That would give us something moderately usable.
@@ -227,21 +242,23 @@ LayoutFileIndex LayoutFileList::addEmptyClass(string const & textclass)
 	// the last parameter to true will suppress a warning message about missing
 	// tex class.
 	LayoutFile * tc = new LayoutFile(textclass, textclass, 
-			"Unknown text class " + textclass, textclass + ".cls", true);
+			"Unknown text class " + textclass, textclass + ".cls", "", true);
 
 	if (!tc->load(tempLayout.absFileName())) {
 		// The only way this happens is because the hardcoded layout file 
 		// above is wrong or stdclass.inc cannot be found. So try again 
-		// without stdclass.inc.
+		// without stdclass.inc and without stdinsets.inc. 
 		ofstream ofs2(tempLayout.toFilesystemEncoding().c_str());
 		ofs2 << "# This layout is automatically generated\n"
 		        "# \\DeclareLaTeXClass{" << textclass << "}\n\n"
 		        "Format " << LAYOUT_FORMAT << "\n"
+		        "Provides stdinsets 1\n"
 		     << layoutpost;
 		ofs2.close();
 		if (!tc->load(tempLayout.absFileName())) {
-			// This can only happen if the hardcoded file above is wrong.
-			LASSERT(false, /* */);
+			// This can only happen if the hardcoded file above is wrong
+			// or there is some weird filesystem error.
+			LATTEST(false); // We will get an empty layout or something.
 		}
 	}
 
@@ -250,8 +267,8 @@ LayoutFileIndex LayoutFileList::addEmptyClass(string const & textclass)
 }
 
 
-LayoutFileIndex 
-	LayoutFileList::addLocalLayout(string const & textclass, string const & path)
+LayoutFileIndex  LayoutFileList::addLocalLayout(
+	string const & textclass, string const & path, string const & oldpath)
 {
 	// FIXME  There is a bug here: 4593
 	//
@@ -260,49 +277,80 @@ LayoutFileIndex
 	// different from textclass
 	string fullName = addName(path, textclass + ".layout");
 	
-	FileName const layout_file(fullName);
-	if (layout_file.exists()) {
-		LYXERR(Debug::TCLASS, "Adding class " << textclass << " from directory " << path);
-		// Read .layout file and get description, real latex classname etc
-		//
-		// This is a C++ version of function processLayoutFile in configure.py,
-		// which uses the following regex
-		//     \Declare(LaTeX|DocBook)Class\s*(\[([^,]*)(,.*)*\])*\s*{(.*)}
-		ifstream ifs(layout_file.toFilesystemEncoding().c_str());
-		static regex const reg("^#\\s*\\\\Declare(LaTeX|DocBook)Class\\s*"
-			"(?:\\[([^,]*)(?:,.*)*\\])*\\s*\\{(.*)\\}\\s*");
-		string line;
-		while (getline(ifs, line)) {
-			// look for the \DeclareXXXClass line
-			smatch sub;
-			if (regex_match(line, sub, reg)) {
-				// returns: whole string, classtype (not used here), class name, description
-				LASSERT(sub.size() == 4, /**/);
-				// now, create a TextClass with description containing path information
-				string class_name(sub.str(2) == "" ? textclass : sub.str(2));
-				string class_prereq(class_name + ".cls");
-				LayoutFile * tmpl = 
-					new LayoutFile(textclass, class_name, textclass, class_prereq, true);
-				//FIXME: The prerequisites are available from the layout file and
-				//       can be extracted from the above regex, but for now this
-				//       field is simply set to class_name + ".cls"
-				// This textclass is added on request so it will definitely be
-				// used. Load it now because other load() calls may fail if they
-				// are called in a context without buffer path information.
-				tmpl->load(path);
-				// There will be only one textclass with this name, even if different
-				// layout files are loaded from different directories.
-				if (haveClass(textclass)) {
-					LYXERR0("Existing textclass " << textclass << " is redefined by " << fullName);
-					delete classmap_[textclass];
-				}
-				classmap_[textclass] = tmpl;
-				return textclass;
-			}
-		}
+	FileName layout_file(fullName);
+	bool moved = false;
+
+	if (!layout_file.exists()) {
+		if (oldpath.empty())
+			return string();
+		// The document has been moved to a different directory.
+		// However, oldpath always points to the right spot, unless
+		// the user also moved the layout file.
+		fullName = addName(oldpath, textclass + ".layout");
+		layout_file.set(fullName);
+		layout_file.refresh();
+		if (!layout_file.exists())
+			return string();
+		moved = true;
 	}
-	// If .layout is not in local directory, or an invalid layout is found, return null
-	return string();
+
+	LYXERR(Debug::TCLASS, "Adding class " << textclass << " from directory " << path);
+	// Read .layout file and get description, real latex classname etc
+	//
+	// This is a C++ version of function processLayoutFile in configure.py,
+	// which uses the following regex
+	//     \Declare(LaTeX|DocBook)Class\s*(\[([^,]*)(,.*)*\])*\s*{(.*)}
+	ifstream ifs(layout_file.toFilesystemEncoding().c_str());
+	static regex const reg("^\\s*#\\s*\\\\Declare(LaTeX|DocBook)Class\\s*"
+		"(?:\\[([^,]*)(?:,.*)*\\])*\\s*\\{(.*)\\}\\s*");
+	static regex const catreg("^\\s*#\\s*\\\\DeclareCategory\\{(.*)\\}\\s*");
+	string line;
+	string class_name;
+	string class_prereq;
+	string category;
+	bool have_declaration = false;
+	while (getline(ifs, line)) {
+		// look for the \DeclareXXXClass line
+		smatch sub;
+		if (regex_match(line, sub, reg)) {
+			// returns: whole string, classtype (not used here), class name, description
+			// LASSERT: Why would this fail?
+			LASSERT(sub.size() == 4, /**/);
+			// now, create a TextClass with description containing path information
+			class_name = (sub.str(2) == "" ? textclass : sub.str(2));
+			class_prereq = class_name + ".cls";
+			have_declaration = true;
+		}
+		else if (regex_match(line, sub, catreg)) {
+			category = sub.str(1);
+		}
+		if (have_declaration && !category.empty())
+			break;
+	}
+
+	if (!have_declaration)
+		return string();
+
+	LayoutFile * tmpl =
+		new LayoutFile(textclass, class_name, textclass, class_prereq, category, true);
+	//FIXME: The prerequisites are available from the layout file and
+	//       can be extracted from the above regex, but for now this
+	//       field is simply set to class_name + ".cls"
+	// This textclass is added on request so it will definitely be
+	// used. Load it now because other load() calls may fail if they
+	// are called in a context without buffer path information.
+	tmpl->load(moved ? oldpath : path);
+	// There will be only one textclass with this name, even if different
+	// layout files are loaded from different directories.
+	if (haveClass(textclass)) {
+		// Unconditionally issuing the warning may be confusing when
+		// saving the document with a different name, as it is exactly
+		// the same textclass that is being re-established.
+		LYXERR(Debug::TCLASS, "Existing textclass " << textclass << " is redefined by " << fullName);
+		delete classmap_[textclass];
+	}
+	classmap_[textclass] = tmpl;
+	return removeExtension(fullName);
 }
 
 
