@@ -50,11 +50,8 @@ using namespace lyx::support;
 namespace lyx {
 
 
-char const lstinline_delimiters[] =
-	"!*()-=+|;:'\"`,<.>/?QWERTYUIOPASDFGHJKLZXCVBNMqwertyuiopasdfghjklzxcvbnm";
-
 InsetListings::InsetListings(Buffer * buf, InsetListingsParams const & par)
-	: InsetCollapsable(buf)
+	: InsetCaptionable(buf,"listing")
 {
 	status_ = par.status();
 }
@@ -72,22 +69,6 @@ Inset::DisplayType InsetListings::display() const
 }
 
 
-void InsetListings::updateBuffer(ParIterator const & it, UpdateType utype)
-{
-	Counters & cnts =
-		buffer().masterBuffer()->params().documentClass().counters();
-	string const saveflt = cnts.current_float();
-
-	// Tell to captions what the current float is
-	cnts.current_float("listing");
-
-	InsetCollapsable::updateBuffer(it, utype);
-
-	//reset afterwards
-	cnts.current_float(saveflt);
-}
-
-
 void InsetListings::write(ostream & os) const
 {
 	os << "listings" << "\n";
@@ -100,7 +81,7 @@ void InsetListings::write(ostream & os) const
 		os << "inline true\n";
 	else
 		os << "inline false\n";
-	InsetCollapsable::write(os);
+	InsetCaptionable::write(os);
 }
 
 
@@ -122,7 +103,7 @@ void InsetListings::read(Lexer & lex)
 			break;
 		}
 	}
-	InsetCollapsable::read(lex);
+	InsetCaptionable::read(lex);
 }
 
 
@@ -141,13 +122,18 @@ void InsetListings::latex(otexstream & os, OutputParams const & runparams) const
 
 	bool encoding_switched = false;
 	Encoding const * const save_enc = runparams.encoding;
+	// The listings package cannot deal with multi-byte-encoded
+	// glyphs, except if full-unicode aware backends
+	// such as XeTeX or LuaTeX are used, and with pLaTeX.
+	bool const multibyte_possible =	runparams.isFullUnicode()
+	    || (buffer().params().bufferFormat() == "platex"
+	        && runparams.encoding->package() == Encoding::japanese);
 
-	if (!runparams.isFullUnicode()
-	    && !runparams.encoding->hasFixedWidth()) {
-		// We need to switch to a singlebyte encoding, since the
-		// listings package cannot deal with multi-byte-encoded
-		// glyphs (not needed with full-unicode aware backends
-		// such as XeTeX).
+	if (!multibyte_possible && !runparams.encoding->hasFixedWidth()) {
+		// We need to switch to a singlebyte encoding, due to
+		// the restrictions of the listings package (see above).
+		// This needs to be consistent with
+		// LaTeXFeatures::getTClassI18nPreamble().
 		Language const * const outer_language =
 			(runparams.local_font != 0) ?
 				runparams.local_font->language()
@@ -156,10 +142,9 @@ void InsetListings::latex(otexstream & os, OutputParams const & runparams) const
 		// language; if not, fall back to latin1.
 		Encoding const * const lstenc =
 			(outer_language->encoding()->hasFixedWidth()) ?
-				outer_language->encoding() 
+				outer_language->encoding()
 				: encodings.fromLyXName("iso8859-1");
-		pair<bool, int> const c = switchEncoding(os.os(),
-				buffer().params(), runparams, *lstenc, true);
+		switchEncoding(os.os(), buffer().params(), runparams, *lstenc, true);
 		runparams.encoding = lstenc;
 		encoding_switched = true;
 	}
@@ -177,22 +162,22 @@ void InsetListings::latex(otexstream & os, OutputParams const & runparams) const
 			// we can only output characters covered by the current
 			// encoding!
 			try {
-				if (runparams.encoding->latexChar(c) == docstring(1, c))
+				if (runparams.encoding->encodable(c))
 					code += c;
 				else if (runparams.dryrun) {
 					code += "<" + _("LyX Warning: ")
 					   + _("uncodable character") + " '";
 					code += docstring(1, c);
 					code += "'>";
- 				} else
+				} else
 					uncodable += c;
 			} catch (EncodingException & /* e */) {
- 				if (runparams.dryrun) {
+				if (runparams.dryrun) {
 					code += "<" + _("LyX Warning: ")
 					   + _("uncodable character") + " '";
 					code += docstring(1, c);
 					code += "'>";
- 				} else
+				} else
 					uncodable += c;
 			}
 		}
@@ -203,18 +188,19 @@ void InsetListings::latex(otexstream & os, OutputParams const & runparams) const
 			code += "\n";
 	}
 	if (isInline) {
-		char const * delimiter = lstinline_delimiters;
-		for (; delimiter != '\0'; ++delimiter)
-			if (!contains(code, *delimiter))
-				break;
+		static const docstring delimiters =
+				from_utf8("!*()-=+|;:'\"`,<.>/?QWERTYUIOPASDFGHJKLZXCVBNMqwertyuiopasdfghjklzxcvbnm");
+
+		size_t pos = delimiters.find_first_not_of(code);
+
 		// This code piece contains all possible special character? !!!
 		// Replace ! with a warning message and use ! as delimiter.
-		if (*delimiter == '\0') {
+		if (pos == string::npos) {
 			docstring delim_error = "<" + _("LyX Warning: ")
 				+ _("no more lstline delimiters available") + ">";
 			code = subst(code, from_ascii("!"), delim_error);
-			delimiter = lstinline_delimiters;
-			if (!runparams.dryrun) {
+			pos = 0;
+			if (!runparams.dryrun && !runparams.silent) {
 				// FIXME: warning should be passed to the error dialog
 				frontend::Alert::warning(_("Running out of delimiters"),
 				_("For inline program listings, one character must be reserved\n"
@@ -224,12 +210,14 @@ void InsetListings::latex(otexstream & os, OutputParams const & runparams) const
 				  "must investigate!"));
 			}
 		}
-		if (param_string.empty())
-			os << "\\lstinline" << *delimiter;
-		else
-			os << "\\lstinline[" << from_utf8(param_string) << "]" << *delimiter;
-                os << code
-                   << *delimiter;
+		docstring const delim(1, delimiters[pos]);
+		os << "\\lstinline";
+		if (!param_string.empty())
+			os << "[" << from_utf8(param_string) << "]";
+		else if (pos >= delimiters.find('Q'))
+			// We need to terminate the command before the delimiter
+			os << " ";
+		os << delim << code << delim;
 	} else {
 		OutputParams rp = runparams;
 		rp.moving_arg = true;
@@ -250,18 +238,27 @@ void InsetListings::latex(otexstream & os, OutputParams const & runparams) const
 
 	if (encoding_switched){
 		// Switch back
-		pair<bool, int> const c = switchEncoding(os.os(),
-				buffer().params(), runparams, *save_enc, true);
+		switchEncoding(os.os(), buffer().params(), runparams, *save_enc, true);
 		runparams.encoding = save_enc;
 	}
 
-	if (!uncodable.empty()) {
+	if (!uncodable.empty() && !runparams.silent) {
 		// issue a warning about omitted characters
 		// FIXME: should be passed to the error dialog
-		frontend::Alert::warning(_("Uncodable characters in listings inset"),
-			bformat(_("The following characters in one of the program listings are\n"
-				  "not representable in the current encoding and have been omitted:\n%1$s."),
-			uncodable));
+		if (!multibyte_possible && !runparams.encoding->hasFixedWidth())
+			frontend::Alert::warning(_("Uncodable characters in listings inset"),
+				bformat(_("The following characters in one of the program listings are\n"
+					  "not representable in the current encoding and have been omitted:\n%1$s.\n"
+					  "This is due to a restriction of the listings package, which does\n"
+					  "not support your encoding '%2$s'.\n"
+					  "Toggling 'Use non-TeX fonts' in Document > Settings...\n"
+					  "might help."),
+				uncodable, _(runparams.encoding->guiName())));
+		else
+			frontend::Alert::warning(_("Uncodable characters in listings inset"),
+				bformat(_("The following characters in one of the program listings are\n"
+					  "not representable in the current encoding and have been omitted:\n%1$s."),
+				uncodable));
 	}
 }
 
@@ -272,21 +269,32 @@ docstring InsetListings::xhtml(XHTMLStream & os, OutputParams const & rp) const
 	XHTMLStream out(ods);
 
 	bool const isInline = params().isInline();
-	if (isInline) 
+	if (isInline)
 		out << html::CompTag("br");
 	else {
 		out << html::StartTag("div", "class='float float-listings'");
 		docstring caption = getCaptionHTML(rp);
 		if (!caption.empty())
-			out << html::StartTag("div", "class='float-caption'") 
+			out << html::StartTag("div", "class='float-caption'")
+			    << XHTMLStream::ESCAPE_NONE
 			    << caption << html::EndTag("div");
 	}
 
-	out << html::StartTag("pre");
+	InsetLayout const & il = getLayout();
+	string const & tag = il.htmltag();
+	string attr = "class ='listings";
+	string const lang = params().getParamValue("language");
+	if (!lang.empty())
+		attr += " " + lang;
+	attr += "'";
+	out << html::StartTag(tag, attr);
 	OutputParams newrp = rp;
 	newrp.html_disable_captions = true;
+	// We don't want to convert dashes here. That's the only conversion we
+	// do for XHTML, so this is safe.
+	newrp.pass_thru = true;
 	docstring def = InsetText::insetAsXHTML(out, newrp, InsetText::JustText);
-	out << html::EndTag("pre");
+	out << html::EndTag(tag);
 
 	if (isInline) {
 		out << html::CompTag("br");
@@ -302,9 +310,9 @@ docstring InsetListings::xhtml(XHTMLStream & os, OutputParams const & rp) const
 }
 
 
-docstring InsetListings::contextMenuName() const
+string InsetListings::contextMenuName() const
 {
-	return from_ascii("context-listings");
+	return "context-listings";
 }
 
 
@@ -313,7 +321,7 @@ void InsetListings::doDispatch(Cursor & cur, FuncRequest & cmd)
 	switch (cmd.action()) {
 
 	case LFUN_INSET_MODIFY: {
-		cur.recordUndoInset(ATOMIC_UNDO, this);
+		cur.recordUndoInset(this);
 		InsetListings::string2params(to_utf8(cmd.argument()), params());
 		break;
 	}
@@ -323,7 +331,7 @@ void InsetListings::doDispatch(Cursor & cur, FuncRequest & cmd)
 		break;
 
 	default:
-		InsetCollapsable::doDispatch(cur, cmd);
+		InsetCaptionable::doDispatch(cur, cmd);
 		break;
 	}
 }
@@ -337,11 +345,15 @@ bool InsetListings::getStatus(Cursor & cur, FuncRequest const & cmd,
 		case LFUN_INSET_DIALOG_UPDATE:
 			status.setEnabled(true);
 			return true;
-		case LFUN_CAPTION_INSERT:
-			status.setEnabled(!params().isInline());
-			return true;
+		case LFUN_CAPTION_INSERT: {
+			// the inset outputs at most one caption
+			if (params().isInline() || getCaptionInset()) {
+				status.setEnabled(false);
+				return true;
+			}
+		}
 		default:
-			return InsetCollapsable::getStatus(cur, cmd, status);
+			return InsetCaptionable::getStatus(cur, cmd, status);
 	}
 }
 
@@ -362,7 +374,7 @@ void InsetListings::validate(LaTeXFeatures & features) const
 	string param_string = params().params();
 	if (param_string.find("\\color") != string::npos)
 		features.require("color");
-	InsetCollapsable::validate(features);
+	InsetCaptionable::validate(features);
 }
 
 
@@ -386,8 +398,12 @@ docstring InsetListings::getCaption(OutputParams const & runparams) const
 	TexRow texrow;
 	odocstringstream ods;
 	otexstream os(ods, texrow);
-	ins->getOptArg(os, runparams);
+	ins->getArgs(os, runparams);
 	ins->getArgument(os, runparams);
+
+	// TODO: The code below should be moved to support, and then the test
+	//       in ../tests should be moved there as well.
+
 	// the caption may contain \label{} but the listings
 	// package prefer caption={}, label={}
 	docstring cap = ods.str();
@@ -402,7 +418,7 @@ docstring InsetListings::getCaption(OutputParams const & runparams) const
 	//
 	// NOTE that } is not allowed in blah2.
 	regex const reg("(.*)\\\\label\\{(.*?)\\}(.*)");
-	string const new_cap("\\1\\3},label={\\2");
+	string const new_cap("$1$3},label={$2");
 	return from_utf8(regex_replace(to_utf8(cap), reg, new_cap));
 }
 
