@@ -4,7 +4,7 @@
  * This file is part of LyX, the document processor.
  * Licence details can be found in the file COPYING.
  *
- * \author Andr Pnitz
+ * \author André Pönitz
  *
  * Full author contact details are available in file CREDITS.
  */
@@ -17,6 +17,7 @@
 #include "InsetMath.h"
 // FIXME: Move to individual insets
 #include "MetricsInfo.h"
+#include "texstream.h"
 
 
 namespace lyx {
@@ -39,10 +40,10 @@ public:
 		wsPreview
 	};
 	///
-	WriteStream(odocstream & os, bool fragile, bool latex, OutputType output,
-		Encoding const * encoding = 0);
+	WriteStream(otexrowstream & os, bool fragile, bool latex, OutputType output,
+				Encoding const * encoding = 0);
 	///
-	explicit WriteStream(odocstream & os);
+	explicit WriteStream(otexrowstream & os);
 	///
 	~WriteStream();
 	///
@@ -54,11 +55,17 @@ public:
 	///
 	OutputType output() const { return output_; }
 	///
-	odocstream & os() { return os_; }
+	otexrowstream & os() { return os_; }
+	///
+	TexRow & texrow() { return os_.texrow(); }
 	///
 	bool & firstitem() { return firstitem_; }
 	///
 	void addlines(unsigned int);
+	/// record whether we can write an immediately following newline char
+	void canBreakLine(bool breakline) { canbreakline_ = breakline; }
+	/// tell whether we can write an immediately following newline char
+	bool canBreakLine() const { return canbreakline_; }
 	/// writes space if next thing is isalpha()
 	void pendingSpace(bool how);
 	/// writes space if next thing is isalpha()
@@ -81,9 +88,18 @@ public:
 	bool asciiOnly() const { return ascii_; }
 	/// LaTeX encoding
 	Encoding const * encoding() const { return encoding_; }
+
+	/// maintains a stack of texrow informations about outer math insets.
+	/// push an entry
+	void pushRowEntry(TexRow::RowEntry entry);
+	/// pop an entry
+	void popRowEntry();
+	/// TexRow::starts the innermost outer math inset
+	/// returns true if the outer row entry will appear at this line
+	bool startOuterRow();
 private:
 	///
-	odocstream & os_;
+	otexrowstream & os_;
 	/// do we have to write \\protect sometimes
 	bool fragile_;
 	/// are we at the beginning of an MathData?
@@ -102,10 +118,14 @@ private:
 	bool locked_;
 	/// should we use only ascii chars when producing latex code?
 	bool ascii_;
+	/// are we allowed to output an immediately following newline?
+	bool canbreakline_;
 	///
 	int line_;
 	///
 	Encoding const * encoding_;
+	///
+	std::vector<TexRow::RowEntry> outer_row_entries_;
 };
 
 ///
@@ -245,9 +265,12 @@ private:
 class MTag {
 public:
 	///
-	MTag(char const * const tag) : tag_(tag) {}
+	MTag(char const * const tag, std::string attr = "") 
+		: tag_(tag), attr_(attr) {}
 	///
 	char const * const tag_;
+	///
+	std::string attr_;
 };
 
 class ETag {
@@ -258,6 +281,14 @@ public:
 	char const * const tag_;
 };
 
+
+/// Throw MathExportException to signal that the attempt to export
+/// some math in the current format did not succeed. E.g., we can't
+/// export xymatrix as MathML, so that will throw, and we'll fall back
+/// to images.
+class MathExportException : public std::exception {};
+
+
 class MathStream {
 public:
 	///
@@ -272,7 +303,17 @@ public:
 	int & tab() { return tab_; }
 	///
 	friend MathStream & operator<<(MathStream &, char const *);
+	///
+	void defer(docstring const &);
+	///
+	void defer(std::string const &);
+	///
+	docstring deferred() const;
+	///
+	bool inText() const { return in_text_; }
 private:
+	///
+	void setTextMode(bool t) { in_text_ = t; }
 	///
 	odocstream & os_;
 	///
@@ -280,7 +321,11 @@ private:
 	///
 	int line_;
 	///
-	char lastchar_;
+	bool in_text_;
+	///
+	odocstringstream deferred_;
+	///
+	friend class SetMode;
 };
 
 ///
@@ -294,11 +339,98 @@ MathStream & operator<<(MathStream &, char const *);
 ///
 MathStream & operator<<(MathStream &, char);
 ///
+MathStream & operator<<(MathStream &, char_type);
+///
 MathStream & operator<<(MathStream &, MTag const &);
 ///
 MathStream & operator<<(MathStream &, ETag const &);
 
 
+/// A simpler version of ModeSpecifier, for MathML
+class SetMode {
+public:
+	///
+	explicit SetMode(MathStream & os, bool text);
+	///
+	~SetMode();
+private:
+	///
+	MathStream & os_;
+	///
+	bool was_text_;
+};
+
+
+class HtmlStream {
+public:
+	///
+	explicit HtmlStream(odocstream & os);
+	///
+	void cr();
+	///
+	odocstream & os() { return os_; }
+	///
+	int line() const { return line_; }
+	///
+	int & tab() { return tab_; }
+	///
+	friend HtmlStream & operator<<(HtmlStream &, char const *);
+	///
+	void defer(docstring const &);
+	///
+	void defer(std::string const &);
+	///
+	docstring deferred() const;
+	///
+	bool inText() const { return in_text_; }
+private:
+	///
+	void setTextMode(bool t) { in_text_ = t; }
+	///
+	odocstream & os_;
+	///
+	int tab_;
+	///
+	int line_;
+	///
+	bool in_text_;
+	///
+	odocstringstream deferred_;
+	///
+	friend class SetHTMLMode;
+};
+
+///
+HtmlStream & operator<<(HtmlStream &, MathAtom const &);
+///
+HtmlStream & operator<<(HtmlStream &, MathData const &);
+///
+HtmlStream & operator<<(HtmlStream &, docstring const &);
+///
+HtmlStream & operator<<(HtmlStream &, char const *);
+///
+HtmlStream & operator<<(HtmlStream &, char);
+///
+HtmlStream & operator<<(HtmlStream &, char_type);
+///
+HtmlStream & operator<<(HtmlStream &, MTag const &);
+///
+HtmlStream & operator<<(HtmlStream &, ETag const &);
+
+
+class SetHTMLMode {
+public:
+	///
+	explicit SetHTMLMode(HtmlStream & os, bool text);
+	///
+	~SetHTMLMode();
+private:
+	///
+	HtmlStream & os_;
+	///
+	bool was_text_;
+};
+
 
 //
 // Debugging
@@ -457,6 +589,9 @@ OctaveStream & operator<<(OctaveStream &, char);
 ///
 OctaveStream & operator<<(OctaveStream &, int);
 
+
+docstring convertDelimToXMLEscape(docstring const & name);
+
 } // namespace lyx
 
 #endif
